use std::{collections::HashMap, sync::Arc};

use serde::Serialize;
use tokio::sync::{Mutex, OnceCell};

use crate::clash_conn_msg::{ClashConnectionMessage, ClashConnectionsWrapper};

#[derive(Clone, Serialize)]
pub struct StatisticsState {
    pub connections: usize,
    pub download_total: u64,
    pub upload_total: u64,
    pub speed_upload: u64,
    pub speed_download: u64,
    pub direct_upload_speed: u64,
    pub direct_download_speed: u64,
    pub proxy_upload_speed: u64,
    pub proxy_download_speed: u64,
    #[serde(skip)]
    connection_map: Arc<Mutex<HashMap<String, ClashConnectionMessage>>>,
}

impl StatisticsState {
    pub fn new() -> Self {
        Self {
            connections: 0,
            download_total: 0,
            upload_total: 0,
            speed_upload: 0,
            speed_download: 0,
            direct_upload_speed: 0,
            direct_download_speed: 0,
            proxy_upload_speed: 0,
            proxy_download_speed: 0,
            connection_map: Arc::new(Mutex::new(HashMap::new())),
        }
    }

    pub async fn update(&mut self, wrapper: &ClashConnectionsWrapper) {
        self.connections = wrapper.connections.len();
        self.update_speed(wrapper.upload_total, wrapper.download_total);
        self.download_total = wrapper.download_total;
        self.upload_total = wrapper.upload_total;

        self.reduce(&wrapper.connections).await;
    }

    pub async fn reduce(&mut self, connections: &Vec<ClashConnectionMessage>) {
        let mut direct_upload_speed = 0;
        let mut direct_download_speed = 0;
        let mut proxy_upload_speed = 0;
        let mut proxy_download_speed = 0;

        let mut map = self.connection_map.lock().await;
        let mut new_map = HashMap::new();

        for connection in connections {
            let id = &connection.id;
            let prev = map.get(id);
            new_map.insert(id.to_string(), connection.clone());

            if prev.is_none() {
                if connection.chains.len() >= 1 {
                    if connection.chains[0] == "DIRECT" {
                        direct_upload_speed += connection.upload;
                        direct_download_speed += connection.download;
                    } else {
                        proxy_upload_speed += connection.upload;
                        proxy_download_speed += connection.download;
                    }
                }
                continue;
            }

            let prev = prev.unwrap();

            if connection.chains.len() >= 1 {
                if connection.chains[0] == "DIRECT" {
                    direct_upload_speed += connection.upload - prev.upload;
                    direct_download_speed += connection.download - prev.download;
                } else {
                    proxy_upload_speed += connection.upload - prev.upload;
                    proxy_download_speed += connection.download - prev.download;
                }
            }
        }

        *map = new_map;

        self.direct_download_speed = direct_download_speed;
        self.direct_upload_speed = direct_upload_speed;
        self.proxy_download_speed = proxy_download_speed;
        self.proxy_upload_speed = proxy_upload_speed;
    }

    pub fn update_speed(&mut self, curr_upload: u64, curr_download: u64) {
        self.speed_upload = curr_upload - self.upload_total;
        self.speed_download = curr_download - self.download_total;
    }
}

pub struct StatisticsManager {
    state: Arc<Mutex<StatisticsState>>,
}

impl StatisticsManager {
    pub async fn global() -> &'static Self {
        static STATISTICS_MANAGER: OnceCell<StatisticsManager> = OnceCell::const_new();

        STATISTICS_MANAGER
            .get_or_init(|| async { StatisticsManager::new() })
            .await
    }

    pub fn new() -> Self {
        Self {
            state: Arc::new(Mutex::new(StatisticsState::new())),
        }
    }

    pub async fn update(&self, wrapper: &ClashConnectionsWrapper) {
        let mut state = self.state.lock().await;
        state.update(wrapper).await;
    }

    pub async fn get_state(&self) -> StatisticsState {
        let state = self.state.lock().await;
        state.clone()
    }
}
